==> amundsen-databuilder-5.2.2.tar.gz: amundsen-databuilder-5.2.2/databuilder/publisher/atlas_csv_publisher.py <==
# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
import logging
from os import listdir
from os.path import isfile, join
from typing import (
    Any, Dict, Iterator, List, Tuple,
)

import pandas
from amundsen_common.utils.atlas import AtlasCommonParams
from apache_atlas.exceptions import AtlasServiceException
from apache_atlas.model.instance import (
    AtlasEntitiesWithExtInfo, AtlasEntity, AtlasObjectId, AtlasRelatedObjectId,
)
from apache_atlas.model.relationship import AtlasRelationship
from pyhocon import ConfigTree

from databuilder.publisher.base_publisher import Publisher
from databuilder.types.atlas import AtlasEntityInitializer
from databuilder.utils.atlas import (
    AtlasSerializedEntityFields, AtlasSerializedEntityOperation, AtlasSerializedRelationshipFields,
)

LOGGER = logging.getLogger(__name__)


class AtlasCSVPublisher(Publisher):
    # atlas client
    ATLAS_CLIENT = 'atlas_client'
    # A directory that contains CSV files for entities
    ENTITY_DIR_PATH = 'entity_files_directory'
    # A directory that contains CSV files for relationships
    RELATIONSHIP_DIR_PATH = 'relationship_files_directory'
    # atlas create entity batch size
    ATLAS_ENTITY_CREATE_BATCH_SIZE = 'batch_size'
    # whether entity types should be registered before data is synced to Atlas
    REGISTER_ENTITY_TYPES = 'register_entity_types'

    def __init__(self) -> None:
        super().__init__()

    def init(self, conf: ConfigTree) -> None:
        self._entity_files = self._list_files(conf, AtlasCSVPublisher.ENTITY_DIR_PATH)
        self._relationship_files = self._list_files(conf, AtlasCSVPublisher.RELATIONSHIP_DIR_PATH)
        self._config = conf
        self._atlas_client = self._config.get(AtlasCSVPublisher.ATLAS_CLIENT)
        self._register_entity_types = self._config.get_bool(AtlasCSVPublisher.REGISTER_ENTITY_TYPES, True)

        if self._register_entity_types:
            LOGGER.info('Registering Atlas Entity Types.')

            try:
                init = AtlasEntityInitializer(self._atlas_client)
                init.create_required_entities()

                LOGGER.info('Registered Atlas Entity Types.')
            except Exception:
                LOGGER.error('Failed to register Atlas Entity Types.', exc_info=True)

    def _list_files(self, conf: ConfigTree, path_key: str) -> List[str]:
        """
        List files from directory
        :param conf:
        :param path_key:
        :return: List of file paths
        """
        if path_key not in conf:
            return []

        path = conf.get_string(path_key)
        return sorted(join(path, f) for f in listdir(path) if isfile(join(path, f)))

    def publish_impl(self) -> None:
        """
        Publishes Entities first and then Relations
        :return:
        """
        LOGGER.info('Creating entities using Entity files: %s', self._entity_files)
        for entity_file in self._entity_files:
            entities_to_create, entities_to_update = self._create_entity_instances(entity_file=entity_file)
            self._sync_entities_to_atlas(entities_to_create)
            self._update_entities(entities_to_update)

        LOGGER.info('Creating relations using relation files: %s', self._relationship_files)
        for relation_file in self._relationship_files:
            self._create_relations(relation_file=relation_file)

    def _update_entities(self, entities_to_update: List[AtlasEntity]) -> None:
        """
        Go over the entities list , create atlas relationships instances and sync them with atlas
        :param entities_to_update:
        :return:
        """
        for entity_to_update in entities_to_update:
            existing_entity = self._atlas_client.entity.get_entity_by_attribute(
                entity_to_update.attributes[AtlasCommonParams.type_name],
                [(AtlasCommonParams.qualified_name, entity_to_update.attributes[AtlasCommonParams.qualified_name])],
            )
            existing_entity.entity.attributes.update(entity_to_update.attributes)
            try:
                self._atlas_client.entity.update_entity(existing_entity)
            except AtlasServiceException:
                LOGGER.error(f'Fail to update entity', exc_info=True)

    def _create_relations(self, relation_file: str) -> None:
        """
        Go over the relation file, create atlas relationships instances and sync them with atlas
        :param relation_file:
        :return:
        """

        with open(relation_file, encoding='utf8') as relation_csv:
            for relation_record in pandas.read_csv(relation_csv, na_filter=False).to_dict(orient='records'):
                relation = self._create_relation(relation_record)
                try:
                    self._atlas_client.relationship.create_relationship(relation)
                except AtlasServiceException:
                    LOGGER.error(f'Fail to create atlas relationship', exc_info=True)
                except Exception as e:
                    LOGGER.error(e)

    def _render_unique_attributes(self, entity_type: str, qualified_name: str) -> Dict[Any, Any]:
        """
        Render uniqueAttributes dict, this struct is needed to identify AtlasObjects
        :param entity_type:
        :param qualified_name:
        :return: rendered uniqueAttributes dict
        """
        return {
            AtlasCommonParams.type_name: entity_type,
            AtlasCommonParams.unique_attributes: {
                AtlasCommonParams.qualified_name: qualified_name,
            },
        }

    def _get_atlas_related_object_id_by_qn(self, entity_type: str, qn: str) -> AtlasRelatedObjectId:
        return AtlasRelatedObjectId(attrs=self._render_unique_attributes(entity_type, qn))

    def _get_atlas_object_id_by_qn(self, entity_type: str, qn: str) -> AtlasObjectId:
        return AtlasObjectId(attrs=self._render_unique_attributes(entity_type, qn))

    def _create_relation(self, relation_dict: Dict[str, str]) -> AtlasRelationship:
        """
        Go over the relation dictionary file and create atlas relationships instances
        :param relation_dict:
        :return:
        """

        relation = AtlasRelationship(
            {AtlasCommonParams.type_name: relation_dict[AtlasSerializedRelationshipFields.relation_type]},
        )
        relation.end1 = self._get_atlas_object_id_by_qn(
            relation_dict[AtlasSerializedRelationshipFields.entity_type_1],
            relation_dict[AtlasSerializedRelationshipFields.qualified_name_1],
        )
        relation.end2 = self._get_atlas_object_id_by_qn(
            relation_dict[AtlasSerializedRelationshipFields.entity_type_2],
            relation_dict[AtlasSerializedRelationshipFields.qualified_name_2],
        )

        return relation

    def _create_entity_instances(self, entity_file: str) -> Tuple[List[AtlasEntity], List[AtlasEntity]]:
        """
        Go over the entities file and try creating instances
        :param entity_file:
        :return:
        """
        entities_to_create = []
        entities_to_update = []
        with open(entity_file, encoding='utf8') as entity_csv:
            for entity_record in pandas.read_csv(entity_csv, na_filter=False).to_dict(orient='records'):
                if entity_record[AtlasSerializedEntityFields.operation] == AtlasSerializedEntityOperation.CREATE:
                    entities_to_create.append(self._create_entity_from_dict(entity_record))
                if entity_record[AtlasSerializedEntityFields.operation] == AtlasSerializedEntityOperation.UPDATE:
                    entities_to_update.append(self._create_entity_from_dict(entity_record))

        return entities_to_create, entities_to_update

    def _extract_entity_relations_details(self, relation_details: str) -> Iterator[Tuple]:
        """
        Generate relation details from relation_attr#related_entity_type#related_qualified_name
        """
        relations = relation_details.split(AtlasSerializedEntityFields.relationships_separator)
        for relation in relations:
            relation_split = relation.split(AtlasSerializedEntityFields.relationships_kv_separator)
            yield relation_split[0], relation_split[1], relation_split[2]

    def _create_entity_from_dict(self, entity_dict: Dict) -> AtlasEntity:
        """
        Create atlas entity instance from dict
        :param entity_dict:
        :return: AtlasEntity
        """
        type_name = {AtlasCommonParams.type_name: entity_dict[AtlasCommonParams.type_name]}
        entity = AtlasEntity(type_name)
        entity.attributes = entity_dict
        relationships = entity_dict.get(AtlasSerializedEntityFields.relationships)
        if relationships:
            relations = {}
            for relation_attr, rel_type, rel_qn in self._extract_entity_relations_details(relationships):
                related_obj = self._get_atlas_related_object_id_by_qn(rel_type, rel_qn)
                relations[relation_attr] = related_obj
            entity.relationshipAttributes = relations
        return entity

    def _chunks(self, lst: List) -> Iterator:
        """
        Yield successive n-sized chunks from lst.
        :param lst:
        :return: chunks generator
        """
        n = self._config.get_int(AtlasCSVPublisher.ATLAS_ENTITY_CREATE_BATCH_SIZE)
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    def _sync_entities_to_atlas(self, entities: List[AtlasEntity]) -> None:
        """
        Sync entities instances with atlas
        :param entities: list of entities
        :return:
        """
        entities_chunks = self._chunks(entities)
        for entity_chunk in entities_chunks:
            LOGGER.info(f'Syncing chunk of {len(entity_chunk)} entities with atlas')
            chunk = AtlasEntitiesWithExtInfo()
            chunk.entities = entity_chunk
            try:
                self._atlas_client.entity.create_entities(chunk)
            except AtlasServiceException:
                LOGGER.error(f'Error during entity syncing', exc_info=True)

    def get_scope(self) -> str:
        return 'publisher.atlas_csv_publisher'

==> amundsen-databuilder-5.2.4.tar.gz: amundsen-databuilder-5.2.4/databuilder/publisher/atlas_csv_publisher.py <==
# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
import logging
from os import listdir
from os.path import isfile, join
from typing import (
    Any, Dict, Iterator, List, Tuple,
)

import pandas
from amundsen_common.utils.atlas import AtlasCommonParams
from apache_atlas.exceptions import AtlasServiceException
from apache_atlas.model.instance import (
    AtlasEntitiesWithExtInfo, AtlasEntity, AtlasObjectId, AtlasRelatedObjectId,
)
from apache_atlas.model.relationship import AtlasRelationship
from pyhocon import ConfigTree

from databuilder.publisher.base_publisher import Publisher
from databuilder.types.atlas import AtlasEntityInitializer
from databuilder.utils.atlas import (
    AtlasSerializedEntityFields, AtlasSerializedEntityOperation, AtlasSerializedRelationshipFields,
)

LOGGER = logging.getLogger(__name__)


class AtlasCSVPublisher(Publisher):
    # atlas client
    ATLAS_CLIENT = 'atlas_client'
    # A directory that contains CSV files for entities
    ENTITY_DIR_PATH = 'entity_files_directory'
    # A directory that contains CSV files for relationships
    RELATIONSHIP_DIR_PATH = 'relationship_files_directory'
    # atlas create entity batch size
    ATLAS_ENTITY_CREATE_BATCH_SIZE = 'batch_size'
    # whether entity types should be registered before data is synced to Atlas
    REGISTER_ENTITY_TYPES = 'register_entity_types'

    def __init__(self) -> None:
        super().__init__()

    def init(self, conf: ConfigTree) -> None:
        self._entity_files = self._list_files(conf, AtlasCSVPublisher.ENTITY_DIR_PATH)
        self._relationship_files = self._list_files(conf, AtlasCSVPublisher.RELATIONSHIP_DIR_PATH)
        self._config = conf
        self._atlas_client = self._config.get(AtlasCSVPublisher.ATLAS_CLIENT)
        self._register_entity_types = self._config.get_bool(AtlasCSVPublisher.REGISTER_ENTITY_TYPES, True)

        if self._register_entity_types:
            LOGGER.info('Registering Atlas Entity Types.')

            try:
                init = AtlasEntityInitializer(self._atlas_client)
                init.create_required_entities()

                LOGGER.info('Registered Atlas Entity Types.')
            except Exception:
                LOGGER.error('Failed to register Atlas Entity Types.', exc_info=True)

    def _list_files(self, conf: ConfigTree, path_key: str) -> List[str]:
        """
        List files from directory
        :param conf:
        :param path_key:
        :return: List of file paths
        """
        if path_key not in conf:
            return []

        path = conf.get_string(path_key)
        return sorted(join(path, f) for f in listdir(path) if isfile(join(path, f)))

    def publish_impl(self) -> None:
        """
        Publishes Entities first and then Relations
        :return:
        """
        LOGGER.info('Creating entities using Entity files: %s', self._entity_files)
        for entity_file in self._entity_files:
            entities_to_create, entities_to_update = self._create_entity_instances(entity_file=entity_file)
            self._sync_entities_to_atlas(entities_to_create)
            self._update_entities(entities_to_update)

        LOGGER.info('Creating relations using relation files: %s', self._relationship_files)
        for relation_file in self._relationship_files:
            self._create_relations(relation_file=relation_file)

    def _update_entities(self, entities_to_update: List[AtlasEntity]) -> None:
        """
        Go over the entities list , create atlas relationships instances and sync them with atlas
        :param entities_to_update:
        :return:
        """
        for entity_to_update in entities_to_update:
            existing_entity = self._atlas_client.entity.get_entity_by_attribute(
                entity_to_update.attributes[AtlasCommonParams.type_name],
                [(AtlasCommonParams.qualified_name, entity_to_update.attributes[AtlasCommonParams.qualified_name])],
            )
            existing_entity.entity.attributes.update(entity_to_update.attributes)
            try:
                self._atlas_client.entity.update_entity(existing_entity)
            except AtlasServiceException:
                LOGGER.error('Fail to update entity', exc_info=True)

    def _create_relations(self, relation_file: str) -> None:
        """
        Go over the relation file, create atlas relationships instances and sync them with atlas
        :param relation_file:
        :return:
        """

        with open(relation_file, encoding='utf8') as relation_csv:
            for relation_record in pandas.read_csv(relation_csv, na_filter=False).to_dict(orient='records'):
                relation = self._create_relation(relation_record)
                try:
                    self._atlas_client.relationship.create_relationship(relation)
                except AtlasServiceException:
                    LOGGER.error('Fail to create atlas relationship', exc_info=True)
                except Exception as e:
                    LOGGER.error(e)

    def _render_unique_attributes(self, entity_type: str, qualified_name: str) -> Dict[Any, Any]:
        """
        Render uniqueAttributes dict, this struct is needed to identify AtlasObjects
        :param entity_type:
        :param qualified_name:
        :return: rendered uniqueAttributes dict
        """
        return {
            AtlasCommonParams.type_name: entity_type,
            AtlasCommonParams.unique_attributes: {
                AtlasCommonParams.qualified_name: qualified_name,
            },
        }

    def _get_atlas_related_object_id_by_qn(self, entity_type: str, qn: str) -> AtlasRelatedObjectId:
        return AtlasRelatedObjectId(attrs=self._render_unique_attributes(entity_type, qn))

    def _get_atlas_object_id_by_qn(self, entity_type: str, qn: str) -> AtlasObjectId:
        return AtlasObjectId(attrs=self._render_unique_attributes(entity_type, qn))

    def _create_relation(self, relation_dict: Dict[str, str]) -> AtlasRelationship:
        """
        Go over the relation dictionary file and create atlas relationships instances
        :param relation_dict:
        :return:
        """

        relation = AtlasRelationship(
            {AtlasCommonParams.type_name: relation_dict[AtlasSerializedRelationshipFields.relation_type]},
        )
        relation.end1 = self._get_atlas_object_id_by_qn(
            relation_dict[AtlasSerializedRelationshipFields.entity_type_1],
            relation_dict[AtlasSerializedRelationshipFields.qualified_name_1],
        )
        relation.end2 = self._get_atlas_object_id_by_qn(
            relation_dict[AtlasSerializedRelationshipFields.entity_type_2],
            relation_dict[AtlasSerializedRelationshipFields.qualified_name_2],
        )

        return relation

    def _create_entity_instances(self, entity_file: str) -> Tuple[List[AtlasEntity], List[AtlasEntity]]:
        """
        Go over the entities file and try creating instances
        :param entity_file:
        :return:
        """
        entities_to_create = []
        entities_to_update = []
        with open(entity_file, encoding='utf8') as entity_csv:
            for entity_record in pandas.read_csv(entity_csv, na_filter=False).to_dict(orient='records'):
                if entity_record[AtlasSerializedEntityFields.operation] == AtlasSerializedEntityOperation.CREATE:
                    entities_to_create.append(self._create_entity_from_dict(entity_record))
                if entity_record[AtlasSerializedEntityFields.operation] == AtlasSerializedEntityOperation.UPDATE:
                    entities_to_update.append(self._create_entity_from_dict(entity_record))

        return entities_to_create, entities_to_update

    def _extract_entity_relations_details(self, relation_details: str) -> Iterator[Tuple]:
        """
        Generate relation details from relation_attr#related_entity_type#related_qualified_name
        """
        relations = relation_details.split(AtlasSerializedEntityFields.relationships_separator)
        for relation in relations:
            relation_split = relation.split(AtlasSerializedEntityFields.relationships_kv_separator)
            yield relation_split[0], relation_split[1], relation_split[2]

    def _create_entity_from_dict(self, entity_dict: Dict) -> AtlasEntity:
        """
        Create atlas entity instance from dict
        :param entity_dict:
        :return: AtlasEntity
        """
        type_name = {AtlasCommonParams.type_name: entity_dict[AtlasCommonParams.type_name]}
        entity = AtlasEntity(type_name)
        entity.attributes = entity_dict
        relationships = entity_dict.get(AtlasSerializedEntityFields.relationships)
        if relationships:
            relations = {}
            for relation_attr, rel_type, rel_qn in self._extract_entity_relations_details(relationships):
                related_obj = self._get_atlas_related_object_id_by_qn(rel_type, rel_qn)
                relations[relation_attr] = related_obj
            entity.relationshipAttributes = relations
        return entity

    def _chunks(self, lst: List) -> Iterator:
        """
        Yield successive n-sized chunks from lst.
        :param lst:
        :return: chunks generator
        """
        n = self._config.get_int(AtlasCSVPublisher.ATLAS_ENTITY_CREATE_BATCH_SIZE)
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    def _sync_entities_to_atlas(self, entities: List[AtlasEntity]) -> None:
        """
        Sync entities instances with atlas
        :param entities: list of entities
        :return:
        """
        entities_chunks = self._chunks(entities)
        for entity_chunk in entities_chunks:
            LOGGER.info(f'Syncing chunk of {len(entity_chunk)} entities with atlas')
            chunk = AtlasEntitiesWithExtInfo()
            chunk.entities = entity_chunk
            try:
                self._atlas_client.entity.create_entities(chunk)
            except AtlasServiceException:
                LOGGER.error('Error during entity syncing', exc_info=True)

    def get_scope(self) -> str:
        return 'publisher.atlas_csv_publisher'
